# pylint: disable=too-many-lines
# pylint: disable=line-too-long
from unittest import mock, TestCase

from pcs_test.tier0.lib.commands.tag.tag_common import fixture_tags_xml
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import (
    TmpFileMock,
    TmpFileCall,
)
from pcs_test.tools.misc import (
    get_test_resource as rc,
    outdent,
)

from pcs import settings
from pcs.common import reports
from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import resource
from pcs.lib.errors import LibraryError


TIMEOUT = 10


fixture_primitive_cib_enabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
        </primitive>
    </resources>
"""
fixture_primitive_cib_enabled_with_meta = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <meta_attributes id="A-meta_attributes" />
        </primitive>
    </resources>
"""
fixture_primitive_cib_disabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        </primitive>
    </resources>
"""
fixture_primitive_status_template = """
    <resources>
        <resource id="A" managed="{managed}" />
    </resources>
"""
fixture_primitive_status_managed = fixture_primitive_status_template.format(
    managed="true"
)
fixture_primitive_status_unmanaged = fixture_primitive_status_template.format(
    managed="false"
)


def get_fixture_two_primitives_cib(
    primitive1_disabled=False,
    primitive1_meta=False,
    primitive2_disabled=False,
    primitive2_meta=False,
):
    parts = [
        """
        <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
    """
    ]
    if primitive1_disabled:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive1_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append(
        """
        </primitive>
        <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
    """
    )
    if primitive2_disabled:
        parts.append(
            """
            <meta_attributes id="B-meta_attributes">
                <nvpair id="B-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive2_meta:
        parts.append("""<meta_attributes id="B-meta_attributes" />""")
    parts.append("""</primitive></resources>""")
    return "".join(parts)


fixture_two_primitives_cib_enabled = get_fixture_two_primitives_cib()
fixture_two_primitives_cib_enabled_with_meta_both = (
    get_fixture_two_primitives_cib(primitive1_meta=True, primitive2_meta=True)
)
fixture_two_primitives_cib_disabled = get_fixture_two_primitives_cib(
    primitive1_disabled=True
)
fixture_two_primitives_cib_disabled_with_meta = get_fixture_two_primitives_cib(
    primitive1_disabled=True, primitive2_meta=True
)
fixture_two_primitives_cib_disabled_both = get_fixture_two_primitives_cib(
    primitive1_disabled=True, primitive2_disabled=True
)
fixture_two_primitives_status_managed = """
    <resources>
        <resource id="A" managed="true" />
        <resource id="B" managed="true" />
    </resources>
"""


def get_fixture_group_cib(
    group_disabled=False,
    group_meta=False,
    primitive1_disabled=False,
    primitive1_meta=False,
    primitive2_disabled=False,
    primitive2_meta=False,
):
    parts = ["""<resources><group id="A">"""]
    if group_disabled:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif group_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append(
        """
        <primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive1_disabled:
        parts.append(
            """
            <meta_attributes id="A1-meta_attributes">
                <nvpair id="A1-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive1_meta:
        parts.append("""<meta_attributes id="A1-meta_attributes" />""")
    parts.append(
        """
        </primitive>
        <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive2_disabled:
        parts.append(
            """
            <meta_attributes id="A2-meta_attributes">
                <nvpair id="A2-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive2_meta:
        parts.append("""<meta_attributes id="A2-meta_attributes" />""")
    parts.append("""</primitive></group></resources>""")
    return "".join(parts)


fixture_group_cib_enabled = get_fixture_group_cib()
fixture_group_cib_enabled_with_meta_group = get_fixture_group_cib(
    group_meta=True
)
fixture_group_cib_enabled_with_meta_primitive = get_fixture_group_cib(
    primitive1_meta=True
)
fixture_group_cib_disabled_group = get_fixture_group_cib(group_disabled=True)
fixture_group_cib_disabled_group_with_meta_primitive = get_fixture_group_cib(
    group_disabled=True, primitive1_meta=True
)
fixture_group_cib_disabled_primitive = get_fixture_group_cib(
    primitive1_disabled=True
)
fixture_group_cib_disabled_primitive_with_meta_group = get_fixture_group_cib(
    group_meta=True, primitive1_disabled=True
)
fixture_group_cib_disabled_both = get_fixture_group_cib(
    group_disabled=True, primitive1_disabled=True
)
# 'disabled' attribute is always set to 'false' even for the test cases where it
# should be set to 'true'. For now it doesn't matter, because pcs doesn't work
# with that attribute. However, if that changes in a future, we may need to
# update the fixtures.
fixture_group_status_template = """
    <resources>
        <group id="A" number_resources="2">
            <resource id="A1" managed="{managed}" />
            <resource id="A2" managed="{managed}" />
        </group>
    </resources>
"""
fixture_group_status_managed = fixture_group_status_template.format(
    managed="true"
)
fixture_group_status_unmanaged = fixture_group_status_template.format(
    managed="false"
)


def get_fixture_clone_cib(
    clone_disabled=False,
    clone_meta=False,
    primitive_disabled=False,
    primitive_meta=False,
):
    parts = ["""<resources><clone id="A-clone">"""]
    if clone_disabled:
        parts.append(
            """
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif clone_meta:
        parts.append("""<meta_attributes id="A-clone-meta_attributes" />""")
    parts.append(
        """<primitive id="A" class="ocf" provider="heartbeat" type="Dummy">"""
    )
    if primitive_disabled:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append("""</primitive></clone></resources>""")
    return "".join(parts)


fixture_clone_cib_enabled = get_fixture_clone_cib()
fixture_clone_cib_enabled_with_meta_both = get_fixture_clone_cib(
    clone_meta=True, primitive_meta=True
)
fixture_clone_cib_enabled_with_meta_clone = get_fixture_clone_cib(
    clone_meta=True
)
fixture_clone_cib_enabled_with_meta_primitive = get_fixture_clone_cib(
    primitive_meta=True
)
fixture_clone_cib_disabled_clone = get_fixture_clone_cib(clone_disabled=True)
fixture_clone_cib_disabled_primitive = get_fixture_clone_cib(
    primitive_disabled=True
)
fixture_clone_cib_disabled_both = get_fixture_clone_cib(
    clone_disabled=True, primitive_disabled=True
)

# 'disabled' attribute is always set to 'false' even for the test cases where it
# should be set to 'true'. For now it doesn't matter, because pcs doesn't work
# with that attribute. However, if that changes in a future, we may need to
# update the fixtures.
fixture_clone_status_template = """
    <resources>
        <clone id="A-clone" managed="{managed}" multi_state="false"
            unique="false"
        >
            <resource id="A" managed="{managed}" />
            <resource id="A" managed="{managed}" />
        </clone>
    </resources>
"""
fixture_clone_status_managed = fixture_clone_status_template.format(
    managed="true"
)
fixture_clone_status_unmanaged = fixture_clone_status_template.format(
    managed="false"
)


def get_fixture_master_cib(
    master_disabled=False,
    master_meta=False,
    primitive_disabled=False,
    primitive_meta=False,
):
    parts = ["""<resources><master id="A-master">"""]
    if master_disabled:
        parts.append(
            """
            <meta_attributes id="A-master-meta_attributes">
                <nvpair id="A-master-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif master_meta:
        parts.append("""<meta_attributes id="A-master-meta_attributes" />""")
    parts.append(
        """<primitive id="A" class="ocf" provider="heartbeat" type="Dummy">"""
    )
    if primitive_disabled:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append("""</primitive></master></resources>""")
    return "".join(parts)


fixture_master_cib_enabled = get_fixture_master_cib()
fixture_master_cib_enabled_with_meta_both = get_fixture_master_cib(
    master_meta=True, primitive_meta=True
)
fixture_master_cib_enabled_with_meta_master = get_fixture_master_cib(
    master_meta=True
)
fixture_master_cib_enabled_with_meta_primitive = get_fixture_master_cib(
    primitive_meta=True
)
fixture_master_cib_disabled_master = get_fixture_master_cib(
    master_disabled=True
)
fixture_master_cib_disabled_primitive = get_fixture_master_cib(
    primitive_disabled=True
)
fixture_master_cib_disabled_both = get_fixture_master_cib(
    master_disabled=True, primitive_disabled=True
)

# 'disabled' attribute is always set to 'false' even for the test cases where it
# should be set to 'true'. For now it doesn't matter, because pcs doesn't work
# with that attribute. However, if that changes in a future, we may need to
# update the fixtures.
fixture_master_status_template = """
    <resources>
        <clone id="A-master" managed="{managed}" multi_state="true"
            unique="false"
        >
            <resource id="A" managed="{managed}" />
            <resource id="A" managed="{managed}" />
        </clone>
    </resources>
"""
fixture_master_status_managed = fixture_master_status_template.format(
    managed="true"
)
fixture_master_status_unmanaged = fixture_master_status_template.format(
    managed="false"
)


def get_fixture_clone_group_cib(
    clone_disabled=False,
    clone_meta=False,
    group_disabled=False,
    group_meta=False,
    primitive1_disabled=False,
    primitive1_meta=False,
    primitive2_disabled=False,
    primitive2_meta=False,
    all_meta=False,
):
    # pylint: disable=too-many-arguments
    parts = ["""<resources><clone id="A-clone">"""]
    if clone_disabled:
        parts.append(
            """
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif clone_meta or all_meta:
        parts.append("""<meta_attributes id="A-clone-meta_attributes" />""")
    parts.append("""<group id="A">""")
    if group_disabled:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif group_meta or all_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append(
        """<primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">"""
    )
    if primitive1_disabled:
        parts.append(
            """
            <meta_attributes id="A1-meta_attributes">
                <nvpair id="A1-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive1_meta or all_meta:
        parts.append("""<meta_attributes id="A1-meta_attributes" />""")
    parts.append(
        """
        </primitive>
        <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive2_disabled:
        parts.append(
            """
            <meta_attributes id="A2-meta_attributes">
                <nvpair id="A2-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive2_meta or all_meta:
        parts.append("""<meta_attributes id="A2-meta_attributes" />""")
    parts.append("""</primitive></group></clone></resources>""")
    return "".join(parts)


fixture_clone_group_cib_enabled = get_fixture_clone_group_cib()
fixture_clone_group_cib_enabled_with_meta_clone = get_fixture_clone_group_cib(
    clone_meta=True
)
fixture_clone_group_cib_enabled_with_meta_group = get_fixture_clone_group_cib(
    group_meta=True
)
fixture_clone_group_cib_enabled_with_meta_primitive = (
    get_fixture_clone_group_cib(primitive1_meta=True)
)
fixture_clone_group_cib_disabled_clone = get_fixture_clone_group_cib(
    clone_disabled=True
)
fixture_clone_group_cib_disabled_group = get_fixture_clone_group_cib(
    group_disabled=True
)
fixture_clone_group_cib_disabled_primitive = get_fixture_clone_group_cib(
    primitive1_disabled=True
)
fixture_clone_group_cib_disabled_primitive_with_meta_clone_group = (
    get_fixture_clone_group_cib(
        clone_meta=True, group_meta=True, primitive1_disabled=True
    )
)
fixture_clone_group_cib_disabled_clone_group_with_meta_primitive = (
    get_fixture_clone_group_cib(
        clone_disabled=True, group_disabled=True, primitive1_meta=True
    )
)
fixture_clone_group_cib_disabled_all = get_fixture_clone_group_cib(
    clone_disabled=True, group_disabled=True, primitive1_disabled=True
)

# 'disabled' attribute is always set to 'false' even for the test cases where it
# should be set to 'true'. For now it doesn't matter, because pcs doesn't work
# with that attribute. However, if that changes in a future, we may need to
# update the fixtures.
fixture_clone_group_status_template = """
    <resources>
        <clone id="A-clone" managed="{managed}" multi_state="false"
            unique="false"
        >
            <group id="A:0" number_resources="2">
                <resource id="A1" managed="{managed}" />
                <resource id="A2" managed="{managed}" />
            </group>
            <group id="A:1" number_resources="2">
                <resource id="A1" managed="{managed}" />
                <resource id="A2" managed="{managed}" />
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_status_managed = fixture_clone_group_status_template.format(
    managed="true"
)
fixture_clone_group_status_unmanaged = (
    fixture_clone_group_status_template.format(managed="false")
)


def get_fixture_bundle_cib(
    bundle_disabled=False,
    bundle_meta=False,
    primitive_disabled=False,
    primitive_meta=False,
):
    parts = ["""<resources><bundle id="A-bundle">"""]
    if bundle_disabled:
        parts.append(
            """
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif bundle_meta:
        parts.append("""<meta_attributes id="A-bundle-meta_attributes" />""")
    parts.append(
        """
        <docker image="pcs:test" />
        <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive_disabled:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        """
        )
    elif primitive_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append("""</primitive></bundle></resources>""")
    return "".join(parts)


fixture_bundle_cib_enabled = get_fixture_bundle_cib()
fixture_bundle_cib_enabled_with_meta_both = get_fixture_bundle_cib(
    bundle_meta=True, primitive_meta=True
)
fixture_bundle_cib_enabled_with_meta_bundle = get_fixture_bundle_cib(
    bundle_meta=True
)
fixture_bundle_cib_enabled_with_meta_primitive = get_fixture_bundle_cib(
    primitive_meta=True
)
fixture_bundle_cib_disabled_primitive = get_fixture_bundle_cib(
    primitive_disabled=True
)
fixture_bundle_cib_disabled_bundle = get_fixture_bundle_cib(
    bundle_disabled=True
)
fixture_bundle_cib_disabled_both = get_fixture_bundle_cib(
    bundle_disabled=True, primitive_disabled=True
)
fixture_bundle_status_template = """
    <resources>
        <bundle id="A-bundle" type="docker" image="pcmktest:http"
            unique="false" managed="{managed}" failed="false"
        >
            <replica id="0">
                <resource id="A" />
            </replica>
            <replica id="1">
                <resource id="A" />
            </replica>
        </bundle>
    </resources>
"""
fixture_bundle_status_managed = fixture_bundle_status_template.format(
    managed="true"
)
fixture_bundle_status_unmanaged = fixture_bundle_status_template.format(
    managed="false"
)


fixture_tag = fixture_tags_xml([("T", ("A", "B"))])


def fixture_report_unmanaged(resource_id):
    return (
        severities.WARNING,
        report_codes.RESOURCE_IS_UNMANAGED,
        {
            "resource_id": resource_id,
        },
        None,
    )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisablePrimitive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_nonexistent_resource(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_enabled)
        self.config.runner.pcmk.load_state(
            resources=fixture_primitive_status_managed
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable(self.env_assist.get_env(), ["B"], False),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("B")]
        )

    def test_nonexistent_resource_in_status(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_enabled
            ).runner.pcmk.load_state(resources=fixture_primitive_status_managed)
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable(self.env_assist.get_env(), ["B"], False)
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found("B"),
            ]
        )

    def test_correct_resource(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_two_primitives_status_managed
            )
            .env.push_cib(resources=fixture_two_primitives_cib_disabled)
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)

    def test_unmanaged(self):
        # The code doesn't care what causes the resource to be unmanaged
        # (cluster property, resource's meta-attribute or whatever). It only
        # checks the cluster state (crm_mon).
        (
            self.config.runner.cib.load(resources=fixture_primitive_cib_enabled)
            .runner.pcmk.load_state(
                resources=fixture_primitive_status_unmanaged
            )
            .env.push_cib(resources=fixture_primitive_cib_disabled)
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports([fixture_report_unmanaged("A")])


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnablePrimitive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_nonexistent_resource(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_disabled)
        self.config.runner.pcmk.load_state()

        self.env_assist.assert_raise_library_error(
            lambda: resource.enable(self.env_assist.get_env(), ["B"], False),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("B")]
        )

    def test_nonexistent_resource_in_status(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_disabled
            ).runner.pcmk.load_state(resources=fixture_primitive_status_managed)
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.enable(self.env_assist.get_env(), ["B"], False)
        )
        self.env_assist.assert_reports([fixture.report_not_found("B")])

    def test_correct_resource(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_disabled_both
            )
            .runner.pcmk.load_state(
                resources=fixture_two_primitives_status_managed
            )
            .env.push_cib(
                resources=fixture_two_primitives_cib_disabled_with_meta
            )
        )
        resource.enable(self.env_assist.get_env(), ["B"], False)

    def test_unmanaged(self):
        # The code doesn't care what causes the resource to be unmanaged
        # (cluster property, resource's meta-attribute or whatever). It only
        # checks the cluster state (crm_mon).
        (
            self.config.runner.cib.load(
                resources=fixture_primitive_cib_disabled
            )
            .runner.pcmk.load_state(
                resources=fixture_primitive_status_unmanaged
            )
            .env.push_cib(resources=fixture_primitive_cib_enabled_with_meta)
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports([fixture_report_unmanaged("A")])


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class MoreResources(TestCase):
    fixture_cib_enabled = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="D" provider="heartbeat" type="Dummy">
            </primitive>
        </resources>
    """
    fixture_cib_disabled = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                <meta_attributes id="B-meta_attributes">
                    <nvpair id="B-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                <meta_attributes id="C-meta_attributes">
                    <nvpair id="C-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="D" provider="heartbeat" type="Dummy">
                <meta_attributes id="D-meta_attributes">
                    <nvpair id="D-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </primitive>
        </resources>
    """
    fixture_status = """
        <resources>
            <resource id="A" managed="true" />
            <resource id="B" managed="false" />
            <resource id="C" managed="true" />
            <resource id="D" managed="false" />
        </resources>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success_enable(self):
        fixture_enabled = """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes" />
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes" />
                </primitive>
                <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                    <meta_attributes id="C-meta_attributes">
                        <nvpair id="C-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="D" provider="heartbeat" type="Dummy">
                    <meta_attributes id="D-meta_attributes" />
                </primitive>
            </resources>
        """
        (
            self.config.runner.cib.load(resources=self.fixture_cib_disabled)
            .runner.pcmk.load_state(resources=self.fixture_status)
            .env.push_cib(resources=fixture_enabled)
        )
        resource.enable(self.env_assist.get_env(), ["A", "B", "D"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("B"),
                fixture_report_unmanaged("D"),
            ]
        )

    def test_success_disable(self):
        fixture_disabled = """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes">
                        <nvpair id="B-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                </primitive>
                <primitive class="ocf" id="D" provider="heartbeat" type="Dummy">
                    <meta_attributes id="D-meta_attributes">
                        <nvpair id="D-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                </primitive>
            </resources>
        """
        (
            self.config.runner.cib.load(resources=self.fixture_cib_enabled)
            .runner.pcmk.load_state(resources=self.fixture_status)
            .env.push_cib(resources=fixture_disabled)
        )
        resource.disable(self.env_assist.get_env(), ["A", "B", "D"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("B"),
                fixture_report_unmanaged("D"),
            ]
        )

    def test_bad_resource_enable(self):
        self.config.runner.cib.load(resources=self.fixture_cib_disabled)
        self.config.runner.pcmk.load_state(resources=self.fixture_status)

        self.env_assist.assert_raise_library_error(
            lambda: resource.enable(
                self.env_assist.get_env(), ["B", "X", "Y", "A"], wait=False
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_resource_or_tag("X"),
                fixture.report_not_resource_or_tag("Y"),
                fixture_report_unmanaged("B"),
            ],
        )

    def test_bad_resource_disable(self):
        self.config.runner.cib.load(resources=self.fixture_cib_enabled)
        self.config.runner.pcmk.load_state(resources=self.fixture_status)

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable(
                self.env_assist.get_env(), ["B", "X", "Y", "A"], wait=False
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_resource_or_tag("X"),
                fixture.report_not_resource_or_tag("Y"),
                fixture_report_unmanaged("B"),
            ],
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class Wait(TestCase):
    fixture_status_running = """
        <resources>
            <resource id="A" managed="true" role="Started">
                <node name="node1" id="1" cached="false"/>
            </resource>
            <resource id="B" managed="true" role="Started">
                <node name="node2" id="1" cached="false"/>
            </resource>
        </resources>
    """
    fixture_status_stopped = """
        <resources>
            <resource id="A" managed="true" role="Stopped">
            </resource>
            <resource id="B" managed="true" role="Stopped">
            </resource>
        </resources>
    """
    fixture_status_mixed = """
        <resources>
            <resource id="A" managed="true" role="Stopped">
            </resource>
            <resource id="B" managed="true" role="Stopped">
            </resource>
        </resources>
    """
    fixture_wait_timeout_error = outdent(
        """\
        Pending actions:
                Action 12: B-node2-stop on node2
        Error performing operation: Timer expired
        """
    ).strip()

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_enable_dont_wait_on_error(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_disabled)
        self.config.runner.pcmk.load_state()

        self.env_assist.assert_raise_library_error(
            lambda: resource.enable(self.env_assist.get_env(), ["B"], TIMEOUT),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("B")]
        )

    def test_disable_dont_wait_on_error(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_enabled)
        self.config.runner.pcmk.load_state()

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable(self.env_assist.get_env(), ["B"], TIMEOUT),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("B")]
        )

    def test_enable_resource_stopped(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=self.fixture_status_stopped)
            .env.push_cib(
                resources=fixture_two_primitives_cib_enabled_with_meta_both,
                wait=TIMEOUT,
            )
            .runner.pcmk.load_state(
                name="",
                resources=self.fixture_status_stopped,
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.enable(
                self.env_assist.get_env(), ["A", "B"], TIMEOUT
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_not_running("A", severities.ERROR),
                fixture.report_resource_not_running("B", severities.ERROR),
            ]
        )

    def test_disable_resource_stopped(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_enabled
            )
            .runner.pcmk.load_state(resources=self.fixture_status_running)
            .env.push_cib(
                resources=fixture_two_primitives_cib_disabled_both, wait=TIMEOUT
            )
            .runner.pcmk.load_state(
                name="",
                resources=self.fixture_status_stopped,
            )
        )

        resource.disable(self.env_assist.get_env(), ["A", "B"], TIMEOUT)
        self.env_assist.assert_reports(
            [
                fixture.report_resource_not_running("A"),
                fixture.report_resource_not_running("B"),
            ]
        )

    def test_enable_resource_running(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=self.fixture_status_stopped)
            .env.push_cib(
                resources=fixture_two_primitives_cib_enabled_with_meta_both,
                wait=TIMEOUT,
            )
            .runner.pcmk.load_state(
                name="",
                resources=self.fixture_status_running,
            )
        )

        resource.enable(self.env_assist.get_env(), ["A", "B"], TIMEOUT)

        self.env_assist.assert_reports(
            [
                fixture.report_resource_running("A", {"Started": ["node1"]}),
                fixture.report_resource_running("B", {"Started": ["node2"]}),
            ]
        )

    def test_disable_resource_running(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_enabled
            )
            .runner.pcmk.load_state(resources=self.fixture_status_running)
            .env.push_cib(
                resources=fixture_two_primitives_cib_disabled_both, wait=TIMEOUT
            )
            .runner.pcmk.load_state(
                name="",
                resources=self.fixture_status_running,
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable(
                self.env_assist.get_env(), ["A", "B"], TIMEOUT
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_running(
                    "A", {"Started": ["node1"]}, severities.ERROR
                ),
                fixture.report_resource_running(
                    "B", {"Started": ["node2"]}, severities.ERROR
                ),
            ]
        )

    def test_enable_wait_timeout(self):
        (
            self.config.runner.cib.load(
                resources=fixture_primitive_cib_disabled
            )
            .runner.pcmk.load_state(resources=self.fixture_status_stopped)
            .env.push_cib(
                resources=fixture_primitive_cib_enabled_with_meta,
                wait=TIMEOUT,
                exception=LibraryError(
                    reports.item.ReportItem.error(
                        reports.messages.WaitForIdleTimedOut(
                            self.fixture_wait_timeout_error
                        )
                    )
                ),
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.enable(self.env_assist.get_env(), ["A"], TIMEOUT),
            [
                fixture.report_wait_for_idle_timed_out(
                    self.fixture_wait_timeout_error
                )
            ],
            expected_in_processor=False,
        )

    def test_disable_wait_timeout(self):
        (
            self.config.runner.cib.load(resources=fixture_primitive_cib_enabled)
            .runner.pcmk.load_state(resources=self.fixture_status_running)
            .env.push_cib(
                resources=fixture_primitive_cib_disabled,
                wait=TIMEOUT,
                exception=LibraryError(
                    reports.item.ReportItem.error(
                        reports.messages.WaitForIdleTimedOut(
                            self.fixture_wait_timeout_error
                        )
                    )
                ),
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable(self.env_assist.get_env(), ["A"], TIMEOUT),
            [
                fixture.report_wait_for_idle_timed_out(
                    self.fixture_wait_timeout_error
                )
            ],
            expected_in_processor=False,
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class WaitClone(TestCase):
    fixture_status_running = """
        <resources>
            <clone id="A-clone" managed="true" multi_state="false" unique="false">
                <resource id="A" managed="true" role="Started">
                    <node name="node1" id="1" cached="false"/>
                </resource>
                <resource id="A" managed="true" role="Started">
                    <node name="node2" id="2" cached="false"/>
                </resource>
            </clone>
        </resources>
    """
    fixture_status_stopped = """
        <resources>
            <clone id="A-clone" managed="true" multi_state="false" unique="false">
                <resource id="A" managed="true" role="Stopped">
                </resource>
                <resource id="A" managed="true" role="Stopped">
                </resource>
            </clone>
        </resources>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_disable_clone(self):
        (
            self.config.runner.cib.load(resources=fixture_clone_cib_enabled)
            .runner.pcmk.load_state(resources=self.fixture_status_running)
            .env.push_cib(
                resources=fixture_clone_cib_disabled_clone, wait=TIMEOUT
            )
            .runner.pcmk.load_state(
                name="",
                resources=self.fixture_status_stopped,
            )
        )

        resource.disable(self.env_assist.get_env(), ["A-clone"], TIMEOUT)
        self.env_assist.assert_reports(
            [
                (
                    severities.INFO,
                    report_codes.RESOURCE_DOES_NOT_RUN,
                    {
                        "resource_id": "A-clone",
                    },
                    None,
                )
            ]
        )

    def test_enable_clone(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_clone
            )
            .runner.pcmk.load_state(resources=self.fixture_status_stopped)
            .env.push_cib(
                resources=fixture_clone_cib_enabled_with_meta_clone,
                wait=TIMEOUT,
            )
            .runner.pcmk.load_state(
                name="",
                resources=self.fixture_status_running,
            )
        )

        resource.enable(self.env_assist.get_env(), ["A-clone"], TIMEOUT)
        self.env_assist.assert_reports(
            [
                (
                    severities.INFO,
                    report_codes.RESOURCE_RUNNING_ON_NODES,
                    {
                        "resource_id": "A-clone",
                        "roles_with_nodes": {"Started": ["node1", "node2"]},
                    },
                    None,
                )
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(resources=fixture_group_cib_enabled)

    def test_primitive(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_group_status_managed
            ).env.push_cib(resources=fixture_group_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A1"], wait=False)

    def test_group(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_group_status_managed
            ).env.push_cib(resources=fixture_group_cib_disabled_group)
        )
        resource.disable(self.env_assist.get_env(), ["A"], wait=False)

    def test_primitive_unmanaged(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_group_status_unmanaged
            ).env.push_cib(resources=fixture_group_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A1"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A1"),
            ]
        )

    def test_group_unmanaged(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_group_status_unmanaged
            ).env.push_cib(resources=fixture_group_cib_disabled_group)
        )
        resource.disable(self.env_assist.get_env(), ["A"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnableGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (
            self.config.runner.cib.load(
                resources=fixture_group_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_group_status_managed)
            .env.push_cib(
                resources=fixture_group_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A1"], wait=False)

    def test_primitive_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_group_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_group_status_managed)
            .env.push_cib(
                resources=fixture_group_cib_disabled_group_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A1"], wait=False)

    def test_group(self):
        (
            self.config.runner.cib.load(
                resources=fixture_group_cib_disabled_group
            )
            .runner.pcmk.load_state(resources=fixture_group_status_managed)
            .env.push_cib(resources=fixture_group_cib_enabled_with_meta_group)
        )
        resource.enable(self.env_assist.get_env(), ["A"], wait=False)

    def test_group_both_disabled(self):
        (
            self.config.runner.cib.load(
                resources=fixture_group_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_group_status_managed)
            .env.push_cib(
                resources=fixture_group_cib_disabled_primitive_with_meta_group
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], wait=False)

    def test_primitive_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_group_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_group_status_unmanaged)
            .env.push_cib(
                resources=fixture_group_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A1"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A1"),
            ]
        )

    def test_group_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_group_cib_disabled_group
            )
            .runner.pcmk.load_state(resources=fixture_group_status_unmanaged)
            .env.push_cib(resources=fixture_group_cib_enabled_with_meta_group)
        )
        resource.enable(self.env_assist.get_env(), ["A"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableClone(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(resources=fixture_clone_cib_enabled)

    def test_primitive(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_clone_status_managed
            ).env.push_cib(resources=fixture_clone_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A"], wait=False)

    def test_clone(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_clone_status_managed
            ).env.push_cib(resources=fixture_clone_cib_disabled_clone)
        )
        resource.disable(self.env_assist.get_env(), ["A-clone"], wait=False)

    def test_primitive_unmanaged(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_clone_status_unmanaged
            ).env.push_cib(resources=fixture_clone_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
            ]
        )

    def test_clone_unmanaged(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_clone_status_unmanaged
            ).env.push_cib(resources=fixture_clone_cib_disabled_clone)
        )
        resource.disable(self.env_assist.get_env(), ["A-clone"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-clone"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnableClone(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_clone_status_managed)
            .env.push_cib(
                resources=fixture_clone_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], wait=False)

    def test_primitive_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_clone_status_managed)
            .env.push_cib(resources=fixture_clone_cib_enabled_with_meta_both)
        )
        resource.enable(self.env_assist.get_env(), ["A"], wait=False)

    def test_clone(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_clone
            )
            .runner.pcmk.load_state(resources=fixture_clone_status_managed)
            .env.push_cib(resources=fixture_clone_cib_enabled_with_meta_clone)
        )
        resource.enable(self.env_assist.get_env(), ["A-clone"], wait=False)

    def test_clone_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_clone_status_managed)
            .env.push_cib(resources=fixture_clone_cib_enabled_with_meta_both)
        )
        resource.enable(self.env_assist.get_env(), ["A-clone"], wait=False)

    def test_primitive_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_clone_status_unmanaged)
            .env.push_cib(
                resources=fixture_clone_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-clone"),
                fixture_report_unmanaged("A"),
            ]
        )

    def test_clone_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_cib_disabled_clone
            )
            .runner.pcmk.load_state(resources=fixture_clone_status_unmanaged)
            .env.push_cib(resources=fixture_clone_cib_enabled_with_meta_clone)
        )
        resource.enable(self.env_assist.get_env(), ["A-clone"], wait=False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-clone"),
                fixture_report_unmanaged("A"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableMaster(TestCase):
    # same as clone, minimum tests in here
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        (
            self.config.runner.cib.load(
                resources=fixture_master_cib_enabled
            ).runner.pcmk.load_state(resources=fixture_master_status_managed)
        )

    def test_primitive(self):
        self.config.env.push_cib(
            resources=fixture_master_cib_disabled_primitive
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)

    def test_master(self):
        self.config.env.push_cib(resources=fixture_master_cib_disabled_master)
        resource.disable(self.env_assist.get_env(), ["A-master"], False)


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnableMaster(TestCase):
    # same as clone, minimum tests in here
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (
            self.config.runner.cib.load(
                resources=fixture_master_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_master_status_managed)
            .env.push_cib(
                resources=fixture_master_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)

    def test_primitive_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_master_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_master_status_managed)
            .env.push_cib(resources=fixture_master_cib_enabled_with_meta_both)
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)

    def test_master(self):
        (
            self.config.runner.cib.load(
                resources=fixture_master_cib_disabled_master
            )
            .runner.pcmk.load_state(resources=fixture_master_status_managed)
            .env.push_cib(resources=fixture_master_cib_enabled_with_meta_master)
        )
        resource.enable(self.env_assist.get_env(), ["A-master"], False)

    def test_master_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_master_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_master_status_managed)
            .env.push_cib(resources=fixture_master_cib_enabled_with_meta_both)
        )
        resource.enable(self.env_assist.get_env(), ["A-master"], False)


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableClonedGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_clone(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(resources=fixture_clone_group_cib_disabled_clone)
        )
        resource.disable(self.env_assist.get_env(), ["A-clone"], False)

    def test_group(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(resources=fixture_clone_group_cib_disabled_group)
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)

    def test_primitive(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(resources=fixture_clone_group_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A1"], False)

    def test_clone_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_unmanaged
            )
            .env.push_cib(resources=fixture_clone_group_cib_disabled_clone)
        )
        resource.disable(self.env_assist.get_env(), ["A-clone"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-clone"),
            ]
        )

    def test_group_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_unmanaged
            )
            .env.push_cib(resources=fixture_clone_group_cib_disabled_group)
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
            ]
        )

    def test_primitive_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_enabled
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_unmanaged
            )
            .env.push_cib(resources=fixture_clone_group_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A1"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A1"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnableClonedGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_clone(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_clone
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_enabled_with_meta_clone
            )
        )
        resource.enable(self.env_assist.get_env(), ["A-clone"], False)

    def test_clone_disabled_all(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_all
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_disabled_primitive_with_meta_clone_group
            )
        )
        resource.enable(self.env_assist.get_env(), ["A-clone"], False)

    def test_group(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_group
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_enabled_with_meta_group
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)

    def test_group_disabled_all(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_all
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_disabled_primitive_with_meta_clone_group
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)

    def test_primitive(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_primitive
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A1"], False)

    def test_primitive_disabled_all(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_all
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_managed
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_disabled_clone_group_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A1"], False)

    def test_clone_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_clone
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_unmanaged
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_enabled_with_meta_clone
            )
        )
        resource.enable(self.env_assist.get_env(), ["A-clone"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-clone"),
                fixture_report_unmanaged("A"),
            ]
        )

    def test_group_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_group
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_unmanaged
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_enabled_with_meta_group
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
                fixture_report_unmanaged("A-clone"),
            ]
        )

    def test_primitive_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_clone_group_cib_disabled_primitive
            )
            .runner.pcmk.load_state(
                resources=fixture_clone_group_status_unmanaged
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A1"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A1"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableBundle(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (
            self.config.runner.cib.load(resources=fixture_bundle_cib_enabled)
            .runner.pcmk.load_state(resources=fixture_bundle_status_managed)
            .env.push_cib(resources=fixture_bundle_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)

    def test_bundle(self):
        (
            self.config.runner.cib.load(resources=fixture_bundle_cib_enabled)
            .runner.pcmk.load_state(resources=fixture_bundle_status_managed)
            .env.push_cib(resources=fixture_bundle_cib_disabled_bundle)
        )
        resource.disable(self.env_assist.get_env(), ["A-bundle"], False)

    def test_primitive_unmanaged(self):
        (
            self.config.runner.cib.load(resources=fixture_bundle_cib_enabled)
            .runner.pcmk.load_state(resources=fixture_bundle_status_unmanaged)
            .env.push_cib(resources=fixture_bundle_cib_disabled_primitive)
        )
        resource.disable(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
            ]
        )

    def test_bundle_unmanaged(self):
        (
            self.config.runner.cib.load(resources=fixture_bundle_cib_enabled)
            .runner.pcmk.load_state(resources=fixture_bundle_status_unmanaged)
            .env.push_cib(resources=fixture_bundle_cib_disabled_bundle)
        )
        resource.disable(self.env_assist.get_env(), ["A-bundle"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-bundle"),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnableBundle(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (
            self.config.runner.cib.load(
                resources=fixture_bundle_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_bundle_status_managed)
            .env.push_cib(
                resources=fixture_bundle_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)

    def test_primitive_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_bundle_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_bundle_status_managed)
            .env.push_cib(resources=fixture_bundle_cib_enabled_with_meta_both)
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)

    def test_bundle(self):
        (
            self.config.runner.cib.load(
                resources=fixture_bundle_cib_disabled_bundle
            )
            .runner.pcmk.load_state(resources=fixture_bundle_status_managed)
            .env.push_cib(resources=fixture_bundle_cib_enabled_with_meta_bundle)
        )
        resource.enable(self.env_assist.get_env(), ["A-bundle"], False)

    def test_bundle_disabled_both(self):
        (
            self.config.runner.cib.load(
                resources=fixture_bundle_cib_disabled_both
            )
            .runner.pcmk.load_state(resources=fixture_bundle_status_managed)
            .env.push_cib(resources=fixture_bundle_cib_enabled_with_meta_both)
        )
        resource.enable(self.env_assist.get_env(), ["A-bundle"], False)

    def test_primitive_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_bundle_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_bundle_status_unmanaged)
            .env.push_cib(
                resources=fixture_bundle_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A"),
                fixture_report_unmanaged("A-bundle"),
            ]
        )

    def test_bundle_unmanaged(self):
        (
            self.config.runner.cib.load(
                resources=fixture_bundle_cib_disabled_primitive
            )
            .runner.pcmk.load_state(resources=fixture_bundle_status_unmanaged)
            .env.push_cib(
                resources=fixture_bundle_cib_enabled_with_meta_primitive
            )
        )
        resource.enable(self.env_assist.get_env(), ["A-bundle"], False)
        self.env_assist.assert_reports(
            [
                fixture_report_unmanaged("A-bundle"),
                fixture_report_unmanaged("A"),
            ]
        )


class DisableSafeFixturesMixin:
    fixture_transitions_both_stopped = """
        <transition_graph>
          <synapse>
            <action_set>
              <rsc_op id="0" operation="stop" on_node="node1">
                <primitive
                    id="A" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
          <synapse>
            <action_set>
              <rsc_op id="1" operation="stop" on_node="node2">
                <primitive
                    id="B" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
        </transition_graph>
    """
    fixture_transitions_one_migrated = """
        <transition_graph>
          <synapse>
            <action_set>
              <rsc_op id="0" operation="stop" on_node="node1">
                <primitive
                    id="A" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
          <synapse>
            <action_set>
              <rsc_op id="1" operation="stop" on_node="node2">
                <primitive
                    id="B" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
          <synapse>
            <action_set>
              <rsc_op id="2" operation="start" on_node="node1">
                <primitive
                    id="B" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
        </transition_graph>
    """
    fixture_transitions_master_demoted = """
        <transition_graph>
          <synapse>
            <action_set>
              <rsc_op id="0" operation="stop" on_node="node1">
                <primitive
                    id="A" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
          <synapse>
            <action_set>
              <rsc_op id="1" operation="demote" on_node="node2">
                <primitive
                    id="B" class="ocf" provider="pacemaker" type="Dummy"
                    long_id="B:0"
                />
              </rsc_op>
            </action_set>
          </synapse>
        </transition_graph>
    """
    fixture_transitions_master_migrated = """
        <transition_graph>
          <synapse>
            <action_set>
              <rsc_op id="0" operation="stop" on_node="node1">
                <primitive
                    id="A" class="ocf" provider="pacemaker" type="Dummy"
                />
              </rsc_op>
            </action_set>
          </synapse>
          <synapse>
            <action_set>
              <rsc_op id="1" operation="demote" on_node="node2">
                <primitive
                    id="B" class="ocf" provider="pacemaker" type="Dummy"
                    long_id="B:0"
                />
              </rsc_op>
            </action_set>
          </synapse>
          <synapse>
            <action_set>
              <rsc_op id="2" operation="promote" on_node="node1">
                <primitive
                    id="B" class="ocf" provider="pacemaker" type="Dummy"
                    long_id="B:1"
                />
              </rsc_op>
            </action_set>
          </synapse>
        </transition_graph>
    """
    fixture_cib_with_master = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            </primitive>
            <master id="B-master">
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                </primitive>
            </master>
        </resources>
    """
    fixture_cib_with_master_primitive_disabled = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </primitive>
            <master id="B-master">
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                </primitive>
            </master>
        </resources>
    """
    fixture_status_with_master_managed = """
        <resources>
            <resource id="A" managed="true" />
            <clone id="B-master" managed="true" multi_state="true"
                unique="false"
            >
                <resource id="B" managed="true" />
                <resource id="B" managed="true" />
            </clone>
        </resources>
    """

    def setUp(self):
        # pylint does not know this will be mixed into TestCase classes
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(test_case=self)
        tmp_file_patcher = mock.patch("pcs.lib.pacemaker.live.get_tmp_file")
        self.addCleanup(tmp_file_patcher.stop)

        self.new_cib_file_name = "new_cib.tmp"
        self.new_cib_content = "<new-cib/>"
        self.transitions_file_name = "transitions.tmp"
        self.transitions_content = "<transitions/>"

        self.tmp_file_mock_obj = TmpFileMock()
        self.addCleanup(self.tmp_file_mock_obj.assert_all_done)
        tmp_file_mock = tmp_file_patcher.start()
        tmp_file_mock.side_effect = (
            self.tmp_file_mock_obj.get_mock_side_effect()
        )

    def fixture_disable_both_resources(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.fixture_transitions_both_stopped,
                ),
            ]
        )
        self.config.runner.cib.load(
            resources=fixture_two_primitives_cib_enabled
        )
        self.config.runner.pcmk.load_state(
            resources=fixture_two_primitives_status_managed
        )

    def fixture_migrate_one_resource(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.fixture_transitions_one_migrated,
                ),
            ]
        )
        self.config.runner.cib.load(
            resources=fixture_two_primitives_cib_enabled
        )
        self.config.runner.pcmk.load_state(
            resources=fixture_two_primitives_status_managed
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableSimulate(DisableSafeFixturesMixin, TestCase):
    def test_not_live(self):
        self.config.env.set_cib_data("<cib />")
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_simulate(
                self.env_assist.get_env(), ["A"], True
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=["CIB"],
                ),
            ],
            expected_in_processor=False,
        )

    def test_nonexistent_resource(self):
        self.config.runner.cib.load()
        self.config.runner.pcmk.load_state()
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_simulate(
                self.env_assist.get_env(), ["A"], True
            ),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("A")],
        )

    def test_success_no_others_stopped(self):
        self.fixture_disable_both_resources()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled_both,
        )

        result = resource.disable_simulate(
            self.env_assist.get_env(), ["A", "B"], True
        )
        self.assertEqual(
            result,
            dict(
                plaintext_simulated_status="simulate output",
                other_affected_resource_list=[],
            ),
        )

    def test_success_others_stopped(self):
        self.fixture_disable_both_resources()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled,
        )

        result = resource.disable_simulate(
            self.env_assist.get_env(), ["A"], True
        )
        self.assertEqual(
            result,
            dict(
                plaintext_simulated_status="simulate output",
                other_affected_resource_list=["B"],
            ),
        )

    def test_success_others_migrated_strict(self):
        self.fixture_migrate_one_resource()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled,
        )
        result = resource.disable_simulate(
            self.env_assist.get_env(), ["A"], True
        )
        self.assertEqual(
            result,
            dict(
                plaintext_simulated_status="simulate output",
                other_affected_resource_list=["B"],
            ),
        )

    def test_success_others_migrated_no_strict(self):
        self.fixture_migrate_one_resource()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled,
        )
        result = resource.disable_simulate(
            self.env_assist.get_env(), ["A"], False
        )
        self.assertEqual(
            result,
            dict(
                plaintext_simulated_status="simulate output",
                other_affected_resource_list=[],
            ),
        )

    def test_success_with_tag_id(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.transitions_content,
                ),
            ]
        )
        self.config.runner.cib.load(
            resources=fixture_two_primitives_cib_enabled,
            tags=fixture_tag,
        )
        self.config.runner.pcmk.load_state(
            resources=fixture_two_primitives_status_managed
        )
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled_both,
        )

        result = resource.disable_simulate(
            self.env_assist.get_env(), ["T"], True
        )
        self.assertEqual(
            result,
            dict(
                plaintext_simulated_status="simulate output",
                other_affected_resource_list=[],
            ),
        )

    def test_simulate_error(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.transitions_content,
                ),
            ]
        )
        self.config.runner.cib.load(resources=fixture_primitive_cib_enabled)
        self.config.runner.pcmk.load_state(
            resources=fixture_primitive_status_managed
        )
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="some stdout",
            stderr="some stderr",
            returncode=1,
            resources=fixture_primitive_cib_disabled,
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_simulate(
                self.env_assist.get_env(), ["A"], True
            ),
            [
                fixture.error(
                    report_codes.CIB_SIMULATE_ERROR,
                    reason="some stderr",
                ),
            ],
            expected_in_processor=False,
        )


class DisableSafeMixin(DisableSafeFixturesMixin):
    def test_not_live(self):
        self.config.env.set_cib_data("<cib />")
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(), ["A"], self.strict, False
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=["CIB"],
                ),
            ],
            expected_in_processor=False,
        )

    def test_nonexistent_resource(self):
        self.config.runner.cib.load()
        self.config.runner.pcmk.load_state()
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(), ["A"], self.strict, False
            ),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("A")]
        )

    def test_simulate_error(self):
        self.fixture_disable_both_resources()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="some stdout",
            stderr="some stderr",
            returncode=1,
            resources=fixture_two_primitives_cib_disabled_both,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(),
                ["A", "B"],
                self.strict,
                False,
            ),
            [
                fixture.error(
                    report_codes.CIB_SIMULATE_ERROR,
                    reason="some stderr",
                ),
            ],
            expected_in_processor=False,
        )

    def test_only_specified_resources_stopped(self):
        self.fixture_disable_both_resources()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled_both,
        )
        self.config.env.push_cib(
            resources=fixture_two_primitives_cib_disabled_both
        )
        resource.disable_safe(
            self.env_assist.get_env(),
            ["A", "B"],
            self.strict,
            False,
        )

    def test_resources_in_tag_stopped(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.fixture_transitions_both_stopped,
                ),
            ]
        )
        self.config.runner.cib.load(
            resources=fixture_two_primitives_cib_enabled,
            tags=fixture_tags_xml([("T1", ("A", "B"))]),
        )
        self.config.runner.pcmk.load_state(
            resources=fixture_two_primitives_status_managed
        )
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled_both,
        )
        self.config.env.push_cib(
            resources=fixture_two_primitives_cib_disabled_both
        )
        resource.disable_safe(
            self.env_assist.get_env(),
            ["T1"],
            self.strict,
            False,
        )

    def test_other_resources_stopped(self):
        self.fixture_disable_both_resources()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(),
                ["A"],
                self.strict,
                False,
            ),
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES,
                    disabled_resource_list=["A"],
                    affected_resource_list=["B"],
                ),
                fixture.info(
                    report_codes.PACEMAKER_SIMULATION_RESULT,
                    plaintext_output="simulate output",
                ),
            ],
        )

    def test_master_demoted(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.fixture_transitions_master_demoted,
                ),
            ]
        )
        self.config.runner.cib.load(resources=self.fixture_cib_with_master)
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_with_master_managed
        )
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=self.fixture_cib_with_master_primitive_disabled,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(),
                ["A"],
                self.strict,
                False,
            ),
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES,
                    disabled_resource_list=["A"],
                    affected_resource_list=["B"],
                ),
                fixture.info(
                    report_codes.PACEMAKER_SIMULATION_RESULT,
                    plaintext_output="simulate output",
                ),
            ]
        )

    def test_wait_success(self):
        self.fixture_disable_both_resources()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled_both,
        )
        self.config.env.push_cib(
            resources=fixture_two_primitives_cib_disabled_both, wait=TIMEOUT
        )
        self.config.runner.pcmk.load_state(
            name="runner.pcmk.load_state_2",
            resources="""
                    <resources>
                        <resource id="A" managed="true" role="Stopped">
                        </resource>
                        <resource id="B" managed="true" role="Stopped">
                        </resource>
                    </resources>
                """,
        )
        resource.disable_safe(
            self.env_assist.get_env(), ["A", "B"], self.strict, TIMEOUT
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_not_running("A"),
                fixture.report_resource_not_running("B"),
            ]
        )

    def test_inner_resources(self):
        cib_xml = """
            <resources>
                <primitive id="A" />
                <clone id="B-clone">
                    <primitive id="B" />
                </clone>
                <master id="C-master">
                    <primitive id="C" />
                </master>
                <group id="D">
                    <primitive id="D1" />
                    <primitive id="D2" />
                </group>
                <clone id="E-clone">
                    <group id="E">
                        <primitive id="E1" />
                        <primitive id="E2" />
                    </group>
                </clone>
                <master id="F-master">
                    <group id="F">
                        <primitive id="F1" />
                        <primitive id="F2" />
                    </group>
                </master>
                <bundle id="G-bundle" />
                <bundle id="H-bundle">
                    <primitive id="H" />
                </bundle>
            </resources>
        """
        status_xml = """
            <resources>
                <resource id="A" managed="true" />
                <clone id="B-clone" managed="true" multi_state="false"
                    unique="false"
                >
                    <resource id="B" managed="true" />
                    <resource id="B" managed="true" />
                </clone>
                <clone id="C-master" managed="true" multi_state="true"
                    unique="false"
                >
                    <resource id="C" managed="true" />
                    <resource id="C" managed="true" />
                </clone>
                <group id="D" number_resources="2">
                    <resource id="D1" managed="true" />
                    <resource id="D2" managed="true" />
                </group>
                <clone id="E-clone" managed="true" multi_state="false"
                    unique="false"
                >
                    <group id="E:0" number_resources="2">
                        <resource id="E1" managed="true" />
                        <resource id="E2" managed="true" />
                    </group>
                    <group id="E:1" number_resources="2">
                        <resource id="E1" managed="true" />
                        <resource id="E2" managed="true" />
                    </group>
                </clone>
                <clone id="F-master" managed="true" multi_state="true"
                    unique="false"
                >
                    <group id="F:0" number_resources="2">
                        <resource id="F1" managed="true" />
                        <resource id="F2" managed="true" />
                    </group>
                    <group id="F:1" number_resources="2">
                        <resource id="F1" managed="true" />
                        <resource id="F2" managed="true" />
                    </group>
                </clone>
                <bundle id="H-bundle" type="docker" image="pcmktest:http"
                    unique="false" managed="true" failed="false"
                >
                    <replica id="0">
                        <resource id="H" />
                    </replica>
                    <replica id="1">
                        <resource id="H" />
                    </replica>
                </bundle>
            </resources>
        """
        synapses = []
        index = 0
        for res_name, is_clone in [
            ("A", False),
            ("B", True),
            ("C", True),
            ("D1", False),
            ("D2", False),
            ("E1", True),
            ("E2", True),
            ("F1", True),
            ("F2", True),
            ("H", False),
        ]:
            if is_clone:
                synapses.append(
                    f"""
                  <synapse>
                    <action_set>
                      <rsc_op id="{index}" operation="stop" on_node="node1">
                        <primitive id="{res_name}" long_id="{res_name}:0" />
                      </rsc_op>
                    </action_set>
                  </synapse>
                  <synapse>
                    <action_set>
                      <rsc_op id="{index + 1}" operation="stop" on_node="node2">
                        <primitive id="{res_name}" long_id="{res_name}:1" />
                      </rsc_op>
                    </action_set>
                  </synapse>
                """
                )
                index += 2
            else:
                synapses.append(
                    f"""
                  <synapse>
                    <action_set>
                      <rsc_op id="{index}" operation="stop" on_node="node1">
                        <primitive id="{res_name}" />
                      </rsc_op>
                    </action_set>
                  </synapse>
                """
                )
                index += 1
        transitions_xml = (
            "<transition_graph>" + "\n".join(synapses) + "</transition_graph>"
        )

        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=transitions_xml,
                ),
            ]
        )
        self.config.runner.cib.load(resources=cib_xml)
        self.config.runner.pcmk.load_state(resources=status_xml)
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources="""
                <resources>
                    <primitive id="A" />
                    <clone id="B-clone">
                        <meta_attributes id="B-clone-meta_attributes">
                            <nvpair name="target-role" value="Stopped"
                                id="B-clone-meta_attributes-target-role"
                            />
                        </meta_attributes>
                        <primitive id="B" />
                    </clone>
                    <master id="C-master">
                        <meta_attributes id="C-master-meta_attributes">
                            <nvpair name="target-role" value="Stopped"
                                id="C-master-meta_attributes-target-role"
                            />
                        </meta_attributes>
                        <primitive id="C" />
                    </master>
                    <group id="D">
                        <meta_attributes id="D-meta_attributes">
                            <nvpair name="target-role" value="Stopped"
                                id="D-meta_attributes-target-role"
                            />
                        </meta_attributes>
                        <primitive id="D1" />
                        <primitive id="D2" />
                    </group>
                    <clone id="E-clone">
                        <meta_attributes id="E-clone-meta_attributes">
                            <nvpair name="target-role" value="Stopped"
                                id="E-clone-meta_attributes-target-role"
                            />
                        </meta_attributes>
                        <group id="E">
                            <primitive id="E1" />
                            <primitive id="E2" />
                        </group>
                    </clone>
                    <master id="F-master">
                        <meta_attributes id="F-master-meta_attributes">
                            <nvpair name="target-role" value="Stopped"
                                id="F-master-meta_attributes-target-role"
                            />
                        </meta_attributes>
                        <group id="F">
                            <primitive id="F1" />
                            <primitive id="F2" />
                        </group>
                    </master>
                    <bundle id="G-bundle" />
                    <bundle id="H-bundle">
                        <meta_attributes id="H-bundle-meta_attributes">
                            <nvpair name="target-role" value="Stopped"
                                id="H-bundle-meta_attributes-target-role"
                            />
                        </meta_attributes>
                        <primitive id="H" />
                    </bundle>
                </resources>
            """,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(),
                ["B-clone", "C-master", "D", "E-clone", "F-master", "H-bundle"],
                self.strict,
                False,
            ),
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES,
                    disabled_resource_list=[
                        "B-clone",
                        "C-master",
                        "D",
                        "E-clone",
                        "F-master",
                        "H-bundle",
                    ],
                    affected_resource_list=["A"],
                ),
                fixture.info(
                    report_codes.PACEMAKER_SIMULATION_RESULT,
                    plaintext_output="simulate output",
                ),
            ],
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableSafe(DisableSafeMixin, TestCase):
    strict = False

    def test_resources_migrated(self):
        self.fixture_migrate_one_resource()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled,
        )
        self.config.env.push_cib(resources=fixture_two_primitives_cib_disabled)
        resource.disable_safe(
            self.env_assist.get_env(),
            ["A"],
            self.strict,
            False,
        )

    def test_master_migrated(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.fixture_transitions_master_migrated,
                ),
            ]
        )
        self.config.runner.cib.load(resources=self.fixture_cib_with_master)
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_with_master_managed
        )
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=self.fixture_cib_with_master_primitive_disabled,
        )
        self.config.env.push_cib(
            resources=self.fixture_cib_with_master_primitive_disabled
        )
        resource.disable_safe(
            self.env_assist.get_env(),
            ["A"],
            self.strict,
            False,
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableSafeStrict(DisableSafeMixin, TestCase):
    strict = True

    def test_resources_migrated(self):
        self.fixture_migrate_one_resource()
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=fixture_two_primitives_cib_disabled,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(),
                ["A"],
                self.strict,
                False,
            ),
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES,
                    disabled_resource_list=["A"],
                    affected_resource_list=["B"],
                ),
                fixture.info(
                    report_codes.PACEMAKER_SIMULATION_RESULT,
                    plaintext_output="simulate output",
                ),
            ]
        )

    def test_master_migrated(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.new_cib_file_name,
                    new_content=self.new_cib_content,
                ),
                TmpFileCall(
                    self.transitions_file_name,
                    new_content=self.fixture_transitions_master_migrated,
                ),
            ]
        )
        self.config.runner.cib.load(resources=self.fixture_cib_with_master)
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_with_master_managed
        )
        self.config.runner.pcmk.simulate_cib(
            self.new_cib_file_name,
            self.transitions_file_name,
            stdout="simulate output",
            resources=self.fixture_cib_with_master_primitive_disabled,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.disable_safe(
                self.env_assist.get_env(),
                ["A"],
                self.strict,
                False,
            ),
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES,
                    disabled_resource_list=["A"],
                    affected_resource_list=["B"],
                ),
                fixture.info(
                    report_codes.PACEMAKER_SIMULATION_RESULT,
                    plaintext_output="simulate output",
                ),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class DisableTags(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_tag_id(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_enabled,
                tags=fixture_tag,
            )
            .runner.pcmk.load_state(
                resources=fixture_two_primitives_status_managed,
            )
            .env.push_cib(resources=fixture_two_primitives_cib_disabled_both)
        )
        resource.disable(self.env_assist.get_env(), ["T"], False)


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class EnableTags(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_tag_id(self):
        (
            self.config.runner.cib.load(
                resources=fixture_two_primitives_cib_disabled_both,
                tags=fixture_tag,
            )
            .runner.pcmk.load_state(
                resources=fixture_two_primitives_status_managed,
            )
            .env.push_cib(
                resources=fixture_two_primitives_cib_enabled_with_meta_both,
            )
        )
        resource.enable(self.env_assist.get_env(), ["T"], False)



import json
import pprint

from pcs_test.tools.command_env.config_http_booth import BoothShortcuts
from pcs_test.tools.command_env.config_http_corosync import CorosyncShortcuts
from pcs_test.tools.command_env.config_http_files import FilesShortcuts
from pcs_test.tools.command_env.config_http_host import HostShortcuts
from pcs_test.tools.command_env.config_http_pcmk import PcmkShortcuts
from pcs_test.tools.command_env.config_http_scsi import ScsiShortcuts
from pcs_test.tools.command_env.config_http_sbd import SbdShortcuts
from pcs_test.tools.command_env.config_http_status import StatusShortcuts
from pcs_test.tools.command_env.mock_node_communicator import (
    place_communication,
    place_requests,
    place_responses,
)
from pcs_test.tools.command_env.mock_node_communicator import (
    place_multinode_call,
)

# pylint: disable=line-too-long


def _mutual_exclusive(param_names, **kwargs):
    entered = {
        key: value
        for key, value in kwargs.items()
        if key in param_names and value is not None
    }
    if len(entered) != 1:
        raise AssertionError(
            "Exactly one of '{0}' must be specified, \nwas specified:\n{1}".format(
                "', '".join(param_names),
                pprint.pformat(entered) if entered else "  nothing",
            )
        )


class HttpConfig:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, call_collection, wrap_helper):
        self.__calls = call_collection

        self.booth = wrap_helper(BoothShortcuts(self.__calls))
        self.corosync = wrap_helper(CorosyncShortcuts(self.__calls))
        self.files = wrap_helper(FilesShortcuts(self.__calls))
        self.host = wrap_helper(HostShortcuts(self.__calls))
        self.pcmk = wrap_helper(PcmkShortcuts(self.__calls))
        self.sbd = wrap_helper(SbdShortcuts(self.__calls))
        self.scsi = wrap_helper(ScsiShortcuts(self.__calls))
        self.status = wrap_helper(StatusShortcuts(self.__calls))

    def add_communication(self, name, communication_list, **kwargs):
        """
        Create a generic call for network communication.
        string name -- key of the call
        list of dict communication_list -- see
            pcs_test.tools.command_env.mock_node_communicator.create_communication
        **kwargs -- see
            pcs_test.tools.command_env.mock_node_communicator.create_communication
        """
        place_communication(self.__calls, name, communication_list, **kwargs)

    def add_requests(self, request_list, name):
        place_requests(self.__calls, name, request_list)

    def start_loop(self, response_list, name):
        place_responses(self.__calls, name, response_list)

    def put_file(
        self,
        communication_list,
        name="http.common.put_file",
        results=None,
        files=None,
        **kwargs,
    ):
        """
        Example:
        config.http.put_file(
            communication_list=[dict(label="node")],
            files={
                "pacemaker_remote authkey": {
                    "type": "pcmk_remote_authkey",
                    "data": base64.b64encode(pcmk_authkey_content),
                    "rewrite_existing": True
                }
            },
            results={
                "pacemaker_remote authkey": {
                    "code": "written",
                    "message": "",
                }
            }
        )
        """

        _mutual_exclusive(["output", "results"], results=results, **kwargs)
        _mutual_exclusive(["files", "param_list"], files=files, **kwargs)

        if results:
            kwargs["output"] = json.dumps({"files": results})

        if files:
            kwargs["param_list"] = [("data_json", json.dumps(files))]

        self.place_multinode_call(
            name,
            communication_list=communication_list,
            action="remote/put_file",
            **kwargs,
        )

    def remove_file(
        self,
        communication_list,
        name="http.common.remove_file",
        results=None,
        files=None,
        **kwargs,
    ):
        """
        Example:
        config.http.remove_file(
            communication_list=[dict(label="node")],
            files={
                "pacemaker_remote authkey": {
                    "type": "pcmk_remote_authkey",
                }
            },
            results={
                "pacemaker_remote authkey": {
                    "code": "deleted",
                    "message": "",
                }
            }
        )
        """

        _mutual_exclusive(["output", "results"], results=results, **kwargs)
        _mutual_exclusive(["files", "param_list"], files=files, **kwargs)

        if results:
            kwargs["output"] = json.dumps({"files": results})

        if files:
            kwargs["param_list"] = [("data_json", json.dumps(files))]

        self.place_multinode_call(
            name,
            communication_list=communication_list,
            action="remote/remove_file",
            **kwargs,
        )

    def manage_services(
        self,
        communication_list,
        name="http.common.manage_services",
        results=None,
        action_map=None,
        **kwargs,
    ):
        """
        Example:
        config.http.manage_services(
            communication_list=[dict(label=label)],
            action_map={
                "pacemaker_remote enable": {
                    "type": "service_command",
                    "service": "pacemaker_remote",
                    "command": "enable",
                },
                "pacemaker_remote start": {
                    "type": "service_command",
                    "service": "pacemaker_remote",
                    "command": "start",
                },
            },
            results={
                "pacemaker_remote enable": {
                    "code": "success",
                    "message": "",
                },
                "pacemaker_remote start": {
                    "code": "success",
                    "message": "",
                }
            }
        )
        """
        _mutual_exclusive(["output", "results"], results=results, **kwargs)
        _mutual_exclusive(
            ["action_map", "param_list"], action_map=action_map, **kwargs
        )

        if results:
            kwargs["output"] = json.dumps({"actions": results})

        if action_map:
            kwargs["param_list"] = [("data_json", json.dumps(action_map))]

        self.place_multinode_call(
            name,
            communication_list=communication_list,
            action="remote/manage_services",
            **kwargs,
        )

    def place_multinode_call(self, *args, **kwargs):
        place_multinode_call(self.__calls, *args, **kwargs)



from textwrap import dedent
from unittest import TestCase

import pcs.lib.corosync.config_facade as lib
from pcs.lib.corosync.config_parser import Parser


def _get_facade(config_text):
    return lib.ConfigFacade(Parser.parse(config_text.encode("utf-8")))


class NeedsStoppedClusterTest(TestCase):
    # pcs.lib.corosync.constants.TRANSPORT_RUNTIME_CHANGE_BANNED_OPTIONS
    # ip_version, netmtu
    _option_list_empty = [
        {"ip_version": ""},
        {"netmtu": ""},
        {"ip_version": "", "netmtu": ""},
    ]
    _option_list_modify = [
        {"ip_version": "ipv6"},
        {"netmtu": "1000"},
        {"ip_version": "ipv4-6", "netmtu": "2000"},
    ]

    def _assert_needs_stoped_cluster(self, config, need_stopped, option_list):
        for options in option_list:
            with self.subTest(options):
                facade = _get_facade(config)
                facade.set_transport_options(options, {}, {})
                self.assertEqual(facade.need_stopped_cluster, need_stopped)

    def test_remove_not_present_options(self):
        self._assert_needs_stoped_cluster("", False, self._option_list_empty)

    def test_remove_present_empty_options(self):
        config = dedent(
            """\
            totem {
                ip_version:
                netmtu:
            }
        """
        )
        self._assert_needs_stoped_cluster(
            config, False, self._option_list_empty
        )

    def test_remove_present_options(self):
        config = dedent(
            """\
            totem {
                ip_version: ipv4
                netmtu: 1500
            }
        """
        )
        self._assert_needs_stoped_cluster(config, True, self._option_list_empty)

    def test_modify_present_options(self):
        config = dedent(
            """\
            totem {
                ip_version: ipv4-6
                netmtu: 1500
            }
        """
        )
        self._assert_needs_stoped_cluster(
            config, True, self._option_list_modify
        )

    def test_not_banned_options(self):
        config = dedent(
            """\
            totem {
                opt1:
                opt2: val2
            }
        """
        )
        not_banned_options = [
            {"opt0": "", "opt1": "", "opt2": ""},
            {"opt0": "val0", "opt1": "val1", "opt2": "val22"},
        ]
        self._assert_needs_stoped_cluster(config, False, not_banned_options)


class SetTransportOptionsKnetMixin:
    knet_transport = None

    def test_add_knet_transport_options(self):
        config = dedent(
            """\
            totem {{{0}
            }}
        """
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "link_mode": "active",
                "knet_pmtud_interval": "1234",
                "ip_version": "ipv4",
            },
            {
                "threshold": "1234",
                "model": "zlib",
                "level": "5",
            },
            {
                "model": "nss",
                "hash": "sha256",
                "cipher": "aes256",
            },
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                totem {{{0}
                    ip_version: ipv4
                    knet_pmtud_interval: 1234
                    link_mode: active
                    knet_compression_level: 5
                    knet_compression_model: zlib
                    knet_compression_threshold: 1234
                    crypto_cipher: aes256
                    crypto_hash: sha256
                    crypto_model: nss
                }}
            """
            ).format(self.knet_transport),
            facade.config.export(),
        )

    def test_remove_knet_transport_options(self):
        config = dedent(
            """\
            totem {{{0}
                ip_version: ipv4
                knet_pmtud_interval: 1234
                link_mode: active
                knet_compression_level: 5
                knet_compression_model: zlib
                knet_compression_threshold: 1234
                crypto_cipher: aes256
                crypto_hash: sha256
                crypto_model: nss
            }}
        """,
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "link_mode": "",
                "knet_pmtud_interval": "",
                "ip_version": "",
            },
            {
                "threshold": "",
                "model": "",
                "level": "",
            },
            {
                "model": "",
                "hash": "",
                "cipher": "",
            },
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            "totem {\n    transport: knet\n}\n" if self.knet_transport else "",
            facade.config.export(),
        )

    def test_modify_knet_transport_options(self):
        config = dedent(
            """\
            totem {{{0}
                ip_version: ipv4
                knet_pmtud_interval: 1234
                link_mode: active
                knet_compression_level: 5
                knet_compression_model: zlib
                knet_compression_threshold: 1234
                crypto_cipher: aes256
                crypto_hash: sha256
                crypto_model: nss
            }}
        """,
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "link_mode": "passive",
                "knet_pmtud_interval": "1000",
                "ip_version": "ipv6",
            },
            {
                "threshold": "1000",
                "model": "lz4",
                "level": "9",
            },
            {
                "model": "openssl",
                "hash": "md5",
                "cipher": "aes128",
            },
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                totem {{{0}
                    ip_version: ipv6
                    knet_pmtud_interval: 1000
                    link_mode: passive
                    knet_compression_level: 9
                    knet_compression_model: lz4
                    knet_compression_threshold: 1000
                    crypto_cipher: aes128
                    crypto_hash: md5
                    crypto_model: openssl
                }}
            """
            ).format(self.knet_transport),
            facade.config.export(),
        )

    def test_add_knet_transport_options_multiple_sections(self):
        config = dedent(
            """\
            totem {{{0}
                knet_pmtud_interval: 1000
                knet_compression_model: zlib
                crypto_cipher: aes256
            }}

            totem {{
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                crypto_hash: sha256
            }}
        """
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "link_mode": "active",
                "ip_version": "ipv6",
            },
            {"level": "5"},
            {"model": "openssl"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                totem {{{0}
                    knet_pmtud_interval: 1000
                    knet_compression_model: zlib
                    crypto_cipher: aes256
                }}

                totem {{
                    knet_pmtud_interval: 2000
                    knet_compression_model: lza4
                    crypto_hash: sha256
                    ip_version: ipv6
                    link_mode: active
                    knet_compression_level: 5
                    crypto_model: openssl
                }}
            """
            ).format(self.knet_transport),
            facade.config.export(),
        )

    def test_remove_knet_transport_options_multiple_sections(self):
        config = dedent(
            """\
            totem {{{0}
                ip_version: ipv6
                knet_pmtud_interval: 1000
                knet_compression_model: zlib
                crypto_cipher: aes256
                ip_version: ipv4
            }}

            totem {{
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv4
                link_mode: active
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                knet_compression_level: 5
                crypto_hash: sha256
                crypto_model: nss
            }}
        """
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "",
                "link_mode": "",
            },
            {"model": ""},
            {"cipher": "", "hash": ""},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                totem {{{0}
                    knet_pmtud_interval: 1000
                }}

                totem {{
                    knet_pmtud_interval: 2000
                    knet_compression_level: 5
                    crypto_model: nss
                }}
            """,
            ).format(self.knet_transport),
            facade.config.export(),
        )

    def test_modify_knet_transport_options_multiple_sections(self):
        config = dedent(
            """\
            totem {{{0}
                ip_version: ipv6
                knet_pmtud_interval: 1000
                link_mode: active
                knet_compression_model: zlib
                knet_compression_level: 5
                knet_compression_threshold: 100
                crypto_cipher: aes256
                crypto_hash: sha512
                ip_version: ipv4
                crypto_model: nss
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv4
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                knet_compression_level: 5
                crypto_hash: sha256
                crypto_model: openssl
            }}
        """
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "ipv4-6",
                "link_mode": "passive",
            },
            {"model": "zlib", "threshold": "100"},
            {"cipher": "aes128", "hash": "md5"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                totem {{{0}
                    knet_pmtud_interval: 1000
                    knet_compression_level: 5
                    crypto_model: nss
                }}

                totem {{
                    ip_version: ipv4-6
                    knet_pmtud_interval: 2000
                    knet_compression_model: zlib
                    knet_compression_level: 5
                    crypto_hash: md5
                    crypto_model: openssl
                    link_mode: passive
                    knet_compression_threshold: 100
                    crypto_cipher: aes128
                }}
            """,
            ).format(self.knet_transport),
            facade.config.export(),
        )

    def test_modify_knet_transport_options_transport_override(self):
        config = dedent(
            """\
            totem {{{0}
                ip_version: ipv6
                knet_pmtud_interval: 1000
                link_mode: active
                knet_compression_model: zlib
                knet_compression_level: 5
                knet_compression_threshold: 100
                crypto_cipher: aes256
                crypto_hash: sha512
                ip_version: ipv4
                crypto_model: nss
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv6
            }}

            totem {{
                transport: udp
                ip_version: ipv4
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                knet_compression_level: 5
                crypto_hash: sha256
                crypto_model: openssl
            }}
        """
        ).format(self.knet_transport)
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "ipv4-6",
                "link_mode": "passive",
            },
            {"model": "zlib", "threshold": "100"},
            {"cipher": "aes128", "hash": "md5"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                totem {{{0}
                    knet_pmtud_interval: 1000
                    knet_compression_model: zlib
                    knet_compression_level: 5
                    knet_compression_threshold: 100
                    crypto_cipher: aes256
                    crypto_hash: sha512
                    crypto_model: nss
                }}

                totem {{
                    transport: udp
                    ip_version: ipv4-6
                    knet_pmtud_interval: 2000
                    knet_compression_model: lza4
                    knet_compression_level: 5
                    crypto_hash: sha256
                    crypto_model: openssl
                    link_mode: passive
                }}
            """,
            ).format(self.knet_transport),
            facade.config.export(),
        )


class SetTransportOptionsDefaultKnetTest(
    SetTransportOptionsKnetMixin, TestCase
):
    knet_transport = ""


class SetTransportOptionsDefinedKnetTest(
    SetTransportOptionsKnetMixin, TestCase
):
    knet_transport = "\n    transport: knet"


class SetTransportOptionsUdpMixin:
    udp_transport = None

    def test_add_udp_transport_options_generic_only(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {"ip_version": "ipv4", "netmtu": "1500"},
            {
                "threshold": "1234",
                "model": "zlib",
                "level": "5",
            },
            {
                "model": "nss",
                "hash": "sha256",
                "cipher": "aes256",
            },
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                    ip_version: ipv4
                    netmtu: 1500
                }}
            """
            ),
            facade.config.export(),
        )

    def test_remove_udp_transport_options_generic_only(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
                ip_version: ipv4
                netmtu: 1500
                knet_pmtud_interval: 1234
                knet_compression_model: zlib
                crypto_cipher: aes256
                crypto_hash: sha256
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "",
                "netmtu": "",
                "knet_pmtud_interval": "",
            },
            {"model": ""},
            {
                "": "",
                "cipher": "",
            },
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                    knet_compression_model: zlib
                    crypto_cipher: aes256
                    crypto_hash: sha256
                }}
            """
            ),
            facade.config.export(),
        )

    def test_modify_udp_transport_options_generic_only(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
                ip_version: ipv4
                netmtu: 1500
                knet_pmtud_interval: 1234
                knet_compression_model: zlib
                crypto_cipher: aes256
                crypto_hash: sha256
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "ipv6",
                "netmtu": "1000",
                "knet_pmtud_interval": "1000",
            },
            {"model": "lz4"},
            {"cipher": "aes128", "hash": "md5"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                    ip_version: ipv6
                    netmtu: 1000
                    knet_pmtud_interval: 1000
                    knet_compression_model: zlib
                    crypto_cipher: aes256
                    crypto_hash: sha256
                }}
            """
            ),
            facade.config.export(),
        )

    def test_add_udp_transport_options_multiple_sections(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
                knet_pmtud_interval: 1000
                knet_compression_model: zlib
                crypto_cipher: aes256
            }}

            totem {{
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                crypto_hash: sha256
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "ipv6",
                "netmtu": "1000",
            },
            {"level": "5"},
            {"model": "openssl"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                    knet_pmtud_interval: 1000
                    knet_compression_model: zlib
                    crypto_cipher: aes256
                }}

                totem {{
                    knet_pmtud_interval: 2000
                    knet_compression_model: lza4
                    crypto_hash: sha256
                    ip_version: ipv6
                    netmtu: 1000
                }}
            """
            ),
            facade.config.export(),
        )

    def test_remove_udp_transport_options_multiple_sections(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
                ip_version: ipv6
                knet_pmtud_interval: 1000
                knet_compression_model: zlib
                crypto_cipher: aes256
                ip_version: ipv4
            }}

            totem {{
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv4
                netmtu: 1000
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                crypto_hash: sha256
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {"ip_version": "", "netmtu": "", "knet_pmtud_interval": ""},
            {"model": ""},
            {"cipher": "", "hash": ""},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                    knet_compression_model: zlib
                    crypto_cipher: aes256
                }}

                totem {{
                    knet_compression_model: lza4
                    crypto_hash: sha256
                }}
            """
            ),
            facade.config.export(),
        )

    def test_modify_udp_transport_options_multiple_sections(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
                ip_version: ipv4
                knet_pmtud_interval: 1000
                knet_compression_model: zlib
                crypto_cipher: aes256
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv6
                netmtu: 1000
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                crypto_hash: sha256
                netmtu: 2000
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "ipv4-6",
                "netmtu": "1500",
                "knet_pmtud_interval": "2000",
            },
            {"model": "lza4"},
            {"cipher": "aes128", "hash": "md5"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                    knet_compression_model: zlib
                    crypto_cipher: aes256
                }}

                totem {{
                    ip_version: ipv4-6
                    netmtu: 1500
                    knet_pmtud_interval: 2000
                    knet_compression_model: lza4
                    crypto_hash: sha256
                }}
            """
            ),
            facade.config.export(),
        )

    def test_modify_udp_transport_options_transport_override(self):
        config = dedent(
            f"""\
            totem {{
                transport: {self.udp_transport}
                ip_version: ipv4
                knet_pmtud_interval: 1000
                knet_compression_model: zlib
                crypto_cipher: aes256
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv6
            }}

            totem {{
                ip_version: ipv6
                transport: knet
                netmtu: 1000
                knet_pmtud_interval: 2000
                knet_compression_model: lza4
                crypto_hash: sha256
                netmtu: 2000
            }}
        """
        )
        facade = _get_facade(config)
        facade.set_transport_options(
            {
                "ip_version": "ipv4-6",
                "netmtu": "1500",
                "knet_pmtud_interval": "2000",
            },
            {"model": "lza4"},
            {"cipher": "aes128", "hash": "md5"},
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                f"""\
                totem {{
                    transport: {self.udp_transport}
                }}

                totem {{
                    ip_version: ipv4-6
                    transport: knet
                    netmtu: 1500
                    knet_pmtud_interval: 2000
                    knet_compression_model: lza4
                    crypto_hash: md5
                    crypto_cipher: aes128
                }}
            """
            ),
            facade.config.export(),
        )


class SetTransportOptionsUdpTest(SetTransportOptionsUdpMixin, TestCase):
    udp_transport = "udp"


class SetTransportOptionsUdpuTest(SetTransportOptionsUdpMixin, TestCase):
    udp_transport = "udpu"


class SetTransportOptionsGeneralTest(TestCase):
    _option_prefix_list = ["", "knet_compression_", "crypto_"]

    _remove_option_params_list = [
        ({"option": ""}, {}, {}),
        ({}, {"option": ""}, {}),
        ({}, {}, {"option": ""}),
    ]

    _add_option_params_list = [
        ({"option": "value"}, {}, {}),
        ({}, {"option": "value"}, {}),
        ({}, {}, {"option": "value"}),
    ]

    _udp_do_not_modify_params_list = [
        ({}, {}, {}),
        ({}, {"option": "value"}, {}),
        ({}, {}, {"option": "value"}),
    ]

    _transport_option_tmplt = dedent(
        """\
        totem {{{{
            transport: {transport}
            {{_prefix}}option: value
        }}}}
    """,
    )

    def _assert_set_transport_options(self, params_list, before, after):
        for idx, params in enumerate(params_list):
            with self.subTest(params=params):
                facade = _get_facade(
                    before.format(_prefix=self._option_prefix_list[idx])
                )
                facade.set_transport_options(*params)
                self.assertFalse(facade.need_qdevice_reload)
                self.assertFalse(facade.need_stopped_cluster)
                self.assertEqual(
                    after.format(_prefix=self._option_prefix_list[idx]),
                    facade.config.export(),
                )

    def test_noop_on_no_options(self):
        self._assert_set_transport_options([({}, {}, {})], "", "")

    def test_add_missing_section(self):
        self._assert_set_transport_options(
            self._add_option_params_list,
            "",
            dedent(
                """\
                totem {{
                    {_prefix}option: value
                }}
            """
            ),
        )

    def test_do_not_add_missing_section(self):
        self._assert_set_transport_options(
            self._remove_option_params_list,
            "",
            "",
        )

    def test_remove_empty_section(self):
        for transport, params in [
            ("", self._remove_option_params_list),
            ("knet", self._remove_option_params_list),
            ("udp", self._remove_option_params_list[0:1]),
            ("udpu", self._remove_option_params_list[0:1]),
        ]:
            with self.subTest(transport=transport, params=params):
                self._assert_set_transport_options(
                    params,
                    dedent(
                        f"""\
                        totem {{{{
                            {{_prefix}}option: value
                        }}}}

                        totem {{{{
                            transport: {transport}
                        }}}}
                    """,
                    ),
                    dedent(
                        f"""\
                        totem {{{{
                            transport: {transport}
                        }}}}
                    """,
                    ),
                )

    def test_do_not_remove_option(self):
        params = self._udp_do_not_modify_params_list
        for transport, params in [("udp", params), ("udpu", params)]:
            with self.subTest(transport=transport, params=params):
                self._assert_set_transport_options(
                    params,
                    self._transport_option_tmplt.format(transport=transport),
                    self._transport_option_tmplt.format(transport=transport),
                )

    def test_modify_option(self):
        params = self._add_option_params_list
        for transport, params in [
            ("", params),
            ("knet", params),
            ("udp", params[0:1]),
            ("udpu", params[0:1]),
        ]:
            with self.subTest(transport=transport, params=params):
                self._assert_set_transport_options(
                    params,
                    dedent(
                        f"""\
                        totem {{{{
                            transport: {transport}
                            {{_prefix}}option: old_value
                        }}}}
                    """,
                    ),
                    self._transport_option_tmplt.format(transport=transport),
                )

    def test_do_not_modify_option(self):
        params = self._udp_do_not_modify_params_list
        for transport, params in [("udp", params), ("udpu", params)]:
            with self.subTest(transport=transport, params=params):
                self._assert_set_transport_options(
                    params,
                    self._transport_option_tmplt.format(transport=transport),
                    self._transport_option_tmplt.format(transport=transport),
                )


class GetOptionsDictMixin:
    _empty_transport_options_template = dedent(
        """
        totem {{
            version: 2
            cluster_name: test99
            transport: {transport}
            ip_version:
            netmtu:
            knet_pmtud_interval:
            link_mode:
            knet_compression_level:
            knet_compression_model:
            knet_compression_threshold:
            crypto_cipher:
            crypto_hash:
            crypto_model:
            block_unlisted_ips:
            consensus:
            downcheck:
            fail_recv_const:
            heartbeat_failures_allowed:
            hold:
            join:
            max_messages:
            max_network_delay:
            merge:
            miss_count_const:
            send_join:
            seqno_unchanged_const:
            token:
            token_coefficient:
            token_retransmit:
            token_retransmits_before_loss_const:
            window_size:
        }}
    """
    )

    _transport_options_template = dedent(
        """
        totem {{
            version: 2
            cluster_name: test99
            transport: {transport}
            ip_version: ipv4-6
            netmtu: 1500
            knet_pmtud_interval: 1234
            link_mode: active
            knet_compression_level: 5
            knet_compression_model: zlib
            knet_compression_threshold: 100
            crypto_cipher: aes256
            crypto_hash: sha256
            crypto_model: nss
            block_unlisted_ips: yes
            consensus: 3600
            downcheck: 1000
            fail_recv_const: 2500
            heartbeat_failures_allowed: 0
            hold: 180
            join: 50
            max_messages: 17
            max_network_delay: 50
            merge: 200
            miss_count_const: 5
            send_join: 0
            seqno_unchanged_const: 30
            token: 3000
            token_coefficient: 650
            token_retransmit: 238
            token_retransmits_before_loss_const: 4
            window_size: 50
        }}
    """
    )

    def _assert_option_dict(self, expected_dict, config):
        facade = _get_facade(config)
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(expected_dict, self.getter(facade))

    def _assert_transport_option_dict(
        self, transport_list, expected_dict, config_template
    ):
        for transport in transport_list:
            with self.subTest(transport=transport):
                self._assert_option_dict(
                    expected_dict, config_template.format(transport=transport)
                )

    def test_empty_config(self):
        self._assert_option_dict({}, "")

    def test_empty_totem_section(self):
        self._assert_option_dict({}, "totem {\n}\n")


class GetTransportOptions(GetOptionsDictMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_transport_options()

    def test_empty_options_knet(self):
        self._assert_transport_option_dict(
            ["", "knet"],
            {"ip_version": "", "knet_pmtud_interval": "", "link_mode": ""},
            self._empty_transport_options_template,
        )

    def test_empty_options_udp(self):
        self._assert_transport_option_dict(
            ["udp", "udpu"],
            {"ip_version": "", "netmtu": ""},
            self._empty_transport_options_template,
        )

    def test_options_knet(self):
        self._assert_transport_option_dict(
            ["", "knet"],
            {
                "ip_version": "ipv4-6",
                "knet_pmtud_interval": "1234",
                "link_mode": "active",
            },
            self._transport_options_template,
        )

    def test_options_udp(self):
        self._assert_transport_option_dict(
            ["udp", "udpu"],
            {"ip_version": "ipv4-6", "netmtu": "1500"},
            self._transport_options_template,
        )


class GetCompressionOptions(GetOptionsDictMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_compression_options()

    def test_empty_compression_options(self):
        self._assert_transport_option_dict(
            ["", "knet", "udp", "udpu"],
            {"level": "", "model": "", "threshold": ""},
            self._empty_transport_options_template,
        )

    def test_compression_options(self):
        self._assert_transport_option_dict(
            ["", "knet", "udp", "udpu"],
            {"level": "5", "model": "zlib", "threshold": "100"},
            self._transport_options_template,
        )


class GetCryptoOptions(GetOptionsDictMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_crypto_options()

    def test_empty_crypto_options(self):
        self._assert_transport_option_dict(
            ["", "knet", "udp", "udpu"],
            {"cipher": "", "hash": "", "model": ""},
            self._empty_transport_options_template,
        )

    def test_crypto_options(self):
        self._assert_transport_option_dict(
            ["", "knet", "udp", "udpu"],
            {"cipher": "aes256", "hash": "sha256", "model": "nss"},
            self._transport_options_template,
        )


FIXTURE_TOTEM_TOKEN_OPTIONS = {
    "block_unlisted_ips": "yes",
    "consensus": "3600",
    "downcheck": "1000",
    "fail_recv_const": "2500",
    "heartbeat_failures_allowed": "0",
    "hold": "180",
    "join": "50",
    "max_messages": "17",
    "max_network_delay": "50",
    "merge": "200",
    "miss_count_const": "5",
    "send_join": "0",
    "seqno_unchanged_const": "30",
    "token": "3000",
    "token_coefficient": "650",
    "token_retransmit": "238",
    "token_retransmits_before_loss_const": "4",
    "window_size": "50",
}


FIXTURE_EMPTY_TOTEM_TOKEN_OPTIONS = {
    "block_unlisted_ips": "",
    "consensus": "",
    "downcheck": "",
    "fail_recv_const": "",
    "heartbeat_failures_allowed": "",
    "hold": "",
    "join": "",
    "max_messages": "",
    "max_network_delay": "",
    "merge": "",
    "miss_count_const": "",
    "send_join": "",
    "seqno_unchanged_const": "",
    "token": "",
    "token_coefficient": "",
    "token_retransmit": "",
    "token_retransmits_before_loss_const": "",
    "window_size": "",
}


class GetTotemOptions(GetOptionsDictMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_totem_options()

    def test_empty_totem_options(self):
        self._assert_transport_option_dict(
            ["", "knet", "udp", "udpu"],
            FIXTURE_EMPTY_TOTEM_TOKEN_OPTIONS,
            self._empty_transport_options_template,
        )

    def test_totem_options(self):
        self._assert_transport_option_dict(
            ["", "knet", "udp", "udpu"],
            FIXTURE_TOTEM_TOKEN_OPTIONS,
            self._transport_options_template,
        )


class SetTotemOptionsTest(TestCase):
    _fixture_totem_token_config = dedent(
        """\
        totem {
            block_unlisted_ips: yes
            consensus: 3600
            downcheck: 1000
            fail_recv_const: 2500
            heartbeat_failures_allowed: 0
            hold: 180
            join: 50
            max_messages: 17
            max_network_delay: 50
            merge: 200
            miss_count_const: 5
            send_join: 0
            seqno_unchanged_const: 30
            token: 3000
            token_coefficient: 650
            token_retransmit: 238
            token_retransmits_before_loss_const: 4
            window_size: 50
        }
    """,
    )

    def _assert_set_totem_options(self, options, before, after):
        facade = _get_facade(before)
        facade.set_totem_options(options)
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(after, facade.config.export())

    def test_add_totem_options(self):
        self._assert_set_totem_options(
            FIXTURE_TOTEM_TOKEN_OPTIONS,
            "",
            self._fixture_totem_token_config,
        )

    def test_remove_totem_options(self):
        self._assert_set_totem_options(
            FIXTURE_EMPTY_TOTEM_TOKEN_OPTIONS,
            self._fixture_totem_token_config,
            "",
        )

    def test_modify_totem_options(self):
        self._assert_set_totem_options(
            {
                "block_unlisted_ips": "no",
                "consensus": "7200",
                "downcheck": "2000",
                "fail_recv_const": "5000",
                "heartbeat_failures_allowed": "1",
                "hold": "360",
                "join": "100",
                "max_messages": "20",
                "max_network_delay": "100",
                "merge": "400",
                "miss_count_const": "10",
                "send_join": "1",
                "seqno_unchanged_const": "60",
                "token": "6000",
                "token_coefficient": "1300",
                "token_retransmit": "476",
                "token_retransmits_before_loss_const": "8",
                "window_size": "100",
            },
            self._fixture_totem_token_config,
            dedent(
                """\
                totem {
                    block_unlisted_ips: no
                    consensus: 7200
                    downcheck: 2000
                    fail_recv_const: 5000
                    heartbeat_failures_allowed: 1
                    hold: 360
                    join: 100
                    max_messages: 20
                    max_network_delay: 100
                    merge: 400
                    miss_count_const: 10
                    send_join: 1
                    seqno_unchanged_const: 60
                    token: 6000
                    token_coefficient: 1300
                    token_retransmit: 476
                    token_retransmits_before_loss_const: 8
                    window_size: 100
                }
            """,
            ),
        )



from typing import List, Optional

from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.str_tools import (
    format_name_value_list,
    indent,
)
from pcs.common.types import (
    CibRuleInEffectStatus,
    CibRuleExpressionType,
)

_in_effect_label_map = {
    CibRuleInEffectStatus.NOT_YET_IN_EFFECT: "not yet in effect",
    CibRuleInEffectStatus.IN_EFFECT: None,
    CibRuleInEffectStatus.EXPIRED: "expired",
}


def get_in_effect_label(rule: CibRuleExpressionDto) -> Optional[str]:
    return _in_effect_label_map.get(rule.in_effect, None)


def rule_expression_dto_to_lines(
    rule_expr: CibRuleExpressionDto, with_ids: bool = False
) -> List[str]:
    if rule_expr.type == CibRuleExpressionType.RULE:
        return _rule_dto_to_lines(rule_expr, with_ids)
    if rule_expr.type == CibRuleExpressionType.DATE_EXPRESSION:
        return _date_dto_to_lines(rule_expr, with_ids)
    return _simple_expr_to_lines(rule_expr, with_ids)


def _rule_dto_to_lines(
    rule_expr: CibRuleExpressionDto, with_ids: bool = False
) -> List[str]:
    in_effect_label = get_in_effect_label(rule_expr)
    heading_parts = [
        "Rule{0}:".format(f" ({in_effect_label})" if in_effect_label else "")
    ]
    heading_parts.extend(
        format_name_value_list(sorted(rule_expr.options.items()))
    )
    if with_ids:
        heading_parts.append(f"(id:{rule_expr.id})")

    lines = []
    for child in rule_expr.expressions:
        lines.extend(rule_expression_dto_to_lines(child, with_ids))

    return [" ".join(heading_parts)] + indent(lines)


def _date_dto_to_lines(
    rule_expr: CibRuleExpressionDto, with_ids: bool = False
) -> List[str]:
    # pylint: disable=too-many-branches
    operation = rule_expr.options.get("operation", None)

    if operation == "date_spec":
        heading_parts = ["Expression:"]
        if with_ids:
            heading_parts.append(f"(id:{rule_expr.id})")
        line_parts = ["Date Spec:"]
        if rule_expr.date_spec:
            line_parts.extend(
                format_name_value_list(
                    sorted(rule_expr.date_spec.options.items())
                )
            )
            if with_ids:
                line_parts.append(f"(id:{rule_expr.date_spec.id})")
        return [" ".join(heading_parts)] + indent([" ".join(line_parts)])

    if operation == "in_range" and rule_expr.duration:
        heading_parts = ["Expression:", "date", "in_range"]
        if "start" in rule_expr.options:
            heading_parts.append(rule_expr.options["start"])
        heading_parts.extend(["to", "duration"])
        if with_ids:
            heading_parts.append(f"(id:{rule_expr.id})")
        lines = [" ".join(heading_parts)]

        line_parts = ["Duration:"]
        line_parts.extend(
            format_name_value_list(sorted(rule_expr.duration.options.items()))
        )
        if with_ids:
            line_parts.append(f"(id:{rule_expr.duration.id})")
        lines.extend(indent([" ".join(line_parts)]))

        return lines

    return _simple_expr_to_lines(rule_expr, with_ids=with_ids)


def _simple_expr_to_lines(
    rule_expr: CibRuleExpressionDto, with_ids: bool = False
) -> List[str]:
    parts = ["Expression:", rule_expr.as_string]
    if with_ids:
        parts.append(f"(id:{rule_expr.id})")
    return [" ".join(parts)]
